"""
This code implements a parser for Linux keyboard maps.
"""

from keymapper.parse.linuxsyms import lookup_symbol,SymbolNotFound
from os.path import exists,join,curdir
from gzip import GzipFile
from keymapper.keymap import Keymap,Modifier, STD,SHIFT,ALT,ALTSHIFT
import os, re

symfile = "/usr/share/misc/keymap.xkeys";
if exists("data/xkeys") and exists("gen_keymap"):
	symfile = "data/xkeys";

xsyms = {}
def read_map():
	f = open(symfile)
	for l in f:
		l = l.strip()
		comment = l.find('#')
		if comment > -1: l = l[:comment]
		l = l.strip()
		if l == "": continue
		s,d = l.split()
		xsyms[s] = eval("0x"+d, {},{})
read_map()
del read_map

defmap="us"
defkey="pc101"

dirs = [
]
def add_dir(d):
	"""Helper to add a directory to the search path"""
	dirs.append(d)

class BadStep(Exception):
	pass
class NoFileFound(Exception):
	pass
class FileProblem(Exception):
	pass
class TooManySyms(Exception):
	pass

class Restart: # Marker
	pass
class UseOld:
	pass
class UseNew:
	pass

def get_file(name):
	"""Find a keymap file"""
	global dirs
	if not dirs:
		dirs=("/usr/lib/X11/xkb/symbols",)
	for end in ("",".gz"):
		for dir in dirs:
			n = os.path.join(dir,name+end)
			if exists(n):
				if n.endswith(".gz"):
					f = GzipFile(n,"rb")
				else:
					f = open(n,"r")
				return f

	raise NoFileFound(name)

m_incl2 = re.compile("^(.+)\\((.+)\\)$")
next_mod = {STD:SHIFT, SHIFT:ALT, ALT:ALTSHIFT, ALTSHIFT:ALT}

class X11Keymap(Keymap):
	"""submap with interesting features for generating X11 maps"""
	def __init__(self,name,sub=None):
		Keymap.__init__(self,name+"("+sub+")")
		self.map_vector = []
		self.steps = []
		self.build_done = False

	def append(self,what):
		self.steps.append(what)

	def build_map(self,to_map=None,keymap=None):
		if to_map is None:
			if self.build_done:
				return
			to_map = self

		if keymap is None:
			keymap={}
		for s in self.steps:
			if isinstance(s,list) or isinstance(s,tuple):
				rep = s[0]
				code = s[1]
				try:
					code = xsyms[code]
				except KeyError:
					continue
				s = s[2:]
				if rep == UseOld and code in keymap:
					continue
				keymap[code] = s

			elif isinstance(s,basestring):
				# include file. Can be "name(sub)".
				m = m_incl2.match(s)
				if m:
					s = (m.group(1),m.group(2))
					if not known.has_key(s):
						parse_file(s[0],s[1])
				else:
					if not known.has_key(s):
						parse_file(s)

				known[s].build_map(to_map,keymap)
			else:
				raise BadStep

		if to_map == self and not self.build_done:
			self.build_done = True
			s = (defmap,defkey)
			if not known.has_key(s):
				parse_file(defmap,defkey)
			known[s].build_map(to_map,keymap)

		for code,keys in keymap.iteritems():
			mod=STD
			for k in keys:
				if k == Restart:
					# mod=STD # heuristic
					continue
				to_map.add(k,code,mod)
				mod=next_mod[mod]
		
known = {}

def parse_file(name, sub=None):
	if sub:
		k = (name,sub)
	else:
		m = m_incl2.match(name)
		if m:
			(name,sub) = (m.group(1),m.group(2))
			k = (name,sub)
		else:
			k = name

	if not known.has_key(k):
		f = get_file(name)
		parser = x11map(x11mapScanner("",file=f,filename=name))

		try:
			parser.map(name)
		except SymbolNotFound,e:
			print >>sys.stderr, "Keysym '%s' not found" % e.args
			print >>sys.stderr, parser._scanner
			raise FileProblem(name)
		except runtime.SyntaxError, e:
			runtime.print_error(e, parser._scanner, max_ctx=1)
			raise FileProblem(name)
		except runtime.NoMoreTokens:
			print >>sys.stderr, 'Could not complete parsing; stopped around here:'
			print >>sys.stderr, parser._scanner
			raise FileProblem(name)

	try:
		map = known[k]
	except KeyError:
		return None
	map.build_map()
	return map


# Begin -- grammar generated by Yapps
import sys, re
from yapps import runtime

class x11mapScanner(runtime.Scanner):
    patterns = [
        ('"LockGroup"', re.compile('LockGroup')),
        ('"\\\\)"', re.compile('\\)')),
        ('"\\\\("', re.compile('\\(')),
        ('"SetMods"', re.compile('SetMods')),
        ('"NoSymbol"', re.compile('NoSymbol')),
        ('"overlay[12]"', re.compile('overlay[12]')),
        ('"type"', re.compile('type')),
        ('"virtualMods"', re.compile('virtualMods')),
        ('"symbols"', re.compile('symbols')),
        ('"key"', re.compile('key')),
        ('"override"', re.compile('override')),
        ('"replace"', re.compile('replace')),
        ('"include"', re.compile('include')),
        ('"alternate_group"', re.compile('alternate_group')),
        ('"modifier_keys"', re.compile('modifier_keys')),
        ('"alphanumeric_keys"', re.compile('alphanumeric_keys')),
        ('"function_keys"', re.compile('function_keys')),
        ('"keypad_keys"', re.compile('keypad_keys')),
        ('"virtual_modifiers"', re.compile('virtual_modifiers')),
        ('","', re.compile(',')),
        ('"modifier_map"', re.compile('modifier_map')),
        ('"="', re.compile('=')),
        ('"\\\\]"', re.compile('\\]')),
        ('"\\\\["', re.compile('\\[')),
        ('"[Nn]ame"', re.compile('[Nn]ame')),
        ('";"', re.compile(';')),
        ('"}"', re.compile('}')),
        ('"{"', re.compile('{')),
        ('"xkb_symbols"', re.compile('xkb_symbols')),
        ('"hidden"', re.compile('hidden')),
        ('"default"', re.compile('default')),
        ('"partial"', re.compile('partial')),
        ('[ \n\r\t]+', re.compile('[ \n\r\t]+')),
        ('(#|!).*', re.compile('(#|!).*')),
        ('\\/\\/.*', re.compile('\\/\\/.*')),
        ('EOF', re.compile('$')),
        ('INCLUDE', re.compile('include')),
        ('SETMODWHAT', re.compile('SetMods\\.[a-zA-Z0-9_]+')),
        ('KEYWHAT', re.compile('key\\.[a-zA-Z0-9_]+')),
        ('HEX', re.compile('0x[a-zA-Z0-9]+')),
        ('UNI', re.compile('U[0-9A-Fa-f]{4}')),
        ('UNI3', re.compile('U[0-9A-Fa-f]{3}')),
        ('SYM', re.compile('[a-zA-Z0-9_]+')),
        ('NUM', re.compile('[-+]?[0-9_]+')),
        ('_STRLITERAL', re.compile('\'([^\\n\'\\\\]|\\\\.)*\'|"([^\\n"\\\\]|\\\\.)*"')),
        ('_KEYLITERAL', re.compile('<[A-Za-z0-9]+>')),
    ]
    def __init__(self, str,*args,**kw):
        runtime.Scanner.__init__(self,None,{'[ \n\r\t]+':None,'(#|!).*':None,'\\/\\/.*':None,},str,*args,**kw)

class x11map(runtime.Parser):
    Context = runtime.Context
    def STRLITERAL(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'STRLITERAL', [])
        _STRLITERAL = self._scan('_STRLITERAL', context=_context)
        return eval(_STRLITERAL,{},{})

    def KEYLITERAL(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'KEYLITERAL', [])
        _KEYLITERAL = self._scan('_KEYLITERAL', context=_context)
        return _KEYLITERAL[1:-1]

    def map(self, name, _parent=None):
        _context = self.Context(_parent, self._scanner, 'map', [name])
        while self._peek('EOF', '"partial"', '"default"', '"hidden"', '"xkb_symbols"', '"keypad_keys"', '"function_keys"', '"alphanumeric_keys"', '"modifier_keys"', '"alternate_group"', context=_context) != 'EOF':
            one_map = self.one_map(name, _context)
        EOF = self._scan('EOF', context=_context)

    def one_map(self, name, _parent=None):
        _context = self.Context(_parent, self._scanner, 'one_map', [name])
        default = False
        hidden = False
        while self._peek('"partial"', '"default"', '"hidden"', '"xkb_symbols"', '"keypad_keys"', '"function_keys"', '"alphanumeric_keys"', '"modifier_keys"', '"alternate_group"', context=_context) in ['"partial"', '"default"', '"hidden"']:
            _token = self._peek('"partial"', '"default"', '"hidden"', context=_context)
            if _token == '"partial"':
                self._scan('"partial"', context=_context)
            elif _token == '"default"':
                self._scan('"default"', context=_context)
                default=True
            else: # == '"hidden"'
                self._scan('"hidden"', context=_context)
                hidden=True
        while self._peek('"xkb_symbols"', '"keypad_keys"', '"function_keys"', '"alphanumeric_keys"', '"modifier_keys"', '"alternate_group"', context=_context) != '"xkb_symbols"':
            mods = self.mods(_context)
        self._scan('"xkb_symbols"', context=_context)
        STRLITERAL = self.STRLITERAL(_context)
        sub=STRLITERAL
        data = X11Keymap(name,sub)
        self._scan('"{"', context=_context)
        while self._peek('"}"', '"include"', '"replace"', '"override"', '"key"', 'SETMODWHAT', '"virtual_modifiers"', '"[Nn]ame"', '"modifier_map"', 'KEYWHAT', context=_context) != '"}"':
            _token = self._peek('"include"', '"replace"', '"override"', '"key"', 'SETMODWHAT', '"virtual_modifiers"', '"[Nn]ame"', '"modifier_map"', 'KEYWHAT', context=_context)
            if _token == '"include"':
                incl = self.incl(data, _context)
            elif _token not in ['SETMODWHAT', '"virtual_modifiers"', '"[Nn]ame"', '"modifier_map"', 'KEYWHAT']:
                entry = self.entry(data, _context)
            elif _token == 'SETMODWHAT':
                setmod = self.setmod(_context)
            elif _token == '"virtual_modifiers"':
                vmod = self.vmod(_context)
            elif _token == '"[Nn]ame"':
                naming = self.naming(_context)
            elif _token == '"modifier_map"':
                modmap = self.modmap(_context)
            else: # == 'KEYWHAT'
                keyset = self.keyset(_context)
        self._scan('"}"', context=_context)
        known[(name,sub)] =  data
        if default: known[name] = data
        self._scan('";"', context=_context)

    def naming(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'naming', [])
        self._scan('"[Nn]ame"', context=_context)
        self._scan('"\\\\["', context=_context)
        SYM = self._scan('SYM', context=_context)
        self._scan('"\\\\]"', context=_context)
        self._scan('"="', context=_context)
        STRLITERAL = self.STRLITERAL(_context)
        self._scan('";"', context=_context)

    def modmap(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'modmap', [])
        self._scan('"modifier_map"', context=_context)
        SYM = self._scan('SYM', context=_context)
        self._scan('"{"', context=_context)
        symkey = self.symkey(_context)
        while self._peek('"}"', '","', 'SYM', '_KEYLITERAL', context=_context) != '"}"':
            if self._peek('","', 'SYM', '_KEYLITERAL', context=_context) == '","':
                self._scan('","', context=_context)
            symkey = self.symkey(_context)
        self._scan('"}"', context=_context)
        self._scan('";"', context=_context)

    def symkey(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'symkey', [])
        _token = self._peek('SYM', '_KEYLITERAL', context=_context)
        if _token == 'SYM':
            SYM = self._scan('SYM', context=_context)
        else: # == '_KEYLITERAL'
            _KEYLITERAL = self._scan('_KEYLITERAL', context=_context)

    def vmod(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'vmod', [])
        self._scan('"virtual_modifiers"', context=_context)
        SYM = self._scan('SYM', context=_context)
        self._scan('";"', context=_context)

    def setmod(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'setmod', [])
        SETMODWHAT = self._scan('SETMODWHAT', context=_context)
        self._scan('"="', context=_context)
        SYM = self._scan('SYM', context=_context)
        self._scan('";"', context=_context)

    def keyset(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'keyset', [])
        KEYWHAT = self._scan('KEYWHAT', context=_context)
        if self._peek('"\\\\["', '"="', context=_context) == '"\\\\["':
            self._scan('"\\\\["', context=_context)
            SYM = self._scan('SYM', context=_context)
            self._scan('"\\\\]"', context=_context)
        self._scan('"="', context=_context)
        STRLITERAL = self.STRLITERAL(_context)
        self._scan('";"', context=_context)

    def mods(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'mods', [])
        _token = self._peek('"keypad_keys"', '"function_keys"', '"alphanumeric_keys"', '"modifier_keys"', '"alternate_group"', context=_context)
        if _token == '"keypad_keys"':
            self._scan('"keypad_keys"', context=_context)
        elif _token == '"function_keys"':
            self._scan('"function_keys"', context=_context)
        elif _token == '"alphanumeric_keys"':
            self._scan('"alphanumeric_keys"', context=_context)
        elif _token == '"modifier_keys"':
            self._scan('"modifier_keys"', context=_context)
        else: # == '"alternate_group"'
            self._scan('"alternate_group"', context=_context)

    def incl(self, data, _parent=None):
        _context = self.Context(_parent, self._scanner, 'incl', [data])
        self._scan('"include"', context=_context)
        STRLITERAL = self.STRLITERAL(_context)
        data.append(STRLITERAL)

    def entry(self, data, _parent=None):
        _context = self.Context(_parent, self._scanner, 'entry', [data])
        rep=UseOld
        if self._peek('"replace"', '"override"', '"key"', context=_context) != '"key"':
            _token = self._peek('"replace"', '"override"', context=_context)
            if _token == '"replace"':
                self._scan('"replace"', context=_context)
            else: # == '"override"'
                self._scan('"override"', context=_context)
            rep=UseNew
        self._scan('"key"', context=_context)
        KEYLITERAL = self.KEYLITERAL(_context)
        k=[rep,KEYLITERAL]
        self._scan('"{"', context=_context)
        stuffgroup = self.stuffgroup(k, _context)
        while self._peek('"}"', '","', '";"', '"{"', '"include"', '"replace"', '"override"', '"key"', 'SETMODWHAT', '"virtual_modifiers"', '"[Nn]ame"', '"modifier_map"', 'KEYWHAT', context=_context) == '","':
            self._scan('","', context=_context)
            k.append(Restart)
            stuffgroup = self.stuffgroup(k, _context)
        self._scan('"}"', context=_context)
        self._scan('";"', context=_context)
        data.append(k)

    def stuffgroup(self, k, _parent=None):
        _context = self.Context(_parent, self._scanner, 'stuffgroup', [k])
        _token = self._peek('"\\\\["', '"symbols"', '"virtualMods"', '"type"', '"overlay[12]"', context=_context)
        if _token != '"\\\\["':
            stuff = self.stuff(_context)
        else: # == '"\\\\["'
            group = self.group(k, _context)

    def stuff(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'stuff', [])
        _token = self._peek('"symbols"', '"virtualMods"', '"type"', '"overlay[12]"', context=_context)
        if _token == '"symbols"':
            stuffsym = self.stuffsym(_context)
        elif _token == '"virtualMods"':
            stuffvirt = self.stuffvirt(_context)
        elif _token == '"type"':
            stufftype = self.stufftype(_context)
        else: # == '"overlay[12]"'
            stuffover = self.stuffover(_context)

    def stuffsym(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'stuffsym', [])
        self._scan('"symbols"', context=_context)
        self._scan('"\\\\["', context=_context)
        SYM = self._scan('SYM', context=_context)
        self._scan('"\\\\]"', context=_context)
        self._scan('"="', context=_context)
        self._scan('"\\\\["', context=_context)
        if self._peek('SYM', '"\\\\]"', '","', context=_context) == 'SYM':
            SYM = self._scan('SYM', context=_context)
            while self._peek('","', '"\\\\]"', context=_context) == '","':
                self._scan('","', context=_context)
                SYM = self._scan('SYM', context=_context)
        self._scan('"\\\\]"', context=_context)

    def stuffvirt(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'stuffvirt', [])
        self._scan('"virtualMods"', context=_context)
        self._scan('"="', context=_context)
        SYM = self._scan('SYM', context=_context)

    def stufftype(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'stufftype', [])
        self._scan('"type"', context=_context)
        if self._peek('"\\\\["', '"="', context=_context) == '"\\\\["':
            self._scan('"\\\\["', context=_context)
            SYM = self._scan('SYM', context=_context)
            self._scan('"\\\\]"', context=_context)
        self._scan('"="', context=_context)
        STRLITERAL = self.STRLITERAL(_context)

    def stuffover(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'stuffover', [])
        self._scan('"overlay[12]"', context=_context)
        self._scan('"="', context=_context)
        KEYLITERAL = self.KEYLITERAL(_context)

    def group(self, k, _parent=None):
        _context = self.Context(_parent, self._scanner, 'group', [k])
        self._scan('"\\\\["', context=_context)
        if self._peek('"\\\\]"', '"NoSymbol"', '"SetMods"', '"LockGroup"', 'UNI', 'UNI3', 'HEX', 'SYM', '","', context=_context) not in ['"\\\\]"', '","']:
            sym = self.sym(_context)
            if sym: k.append(sym)
            while self._peek('","', '"NoSymbol"', '"SetMods"', '"LockGroup"', 'UNI', 'UNI3', 'HEX', 'SYM', '"\\\\]"', context=_context) != '"\\\\]"':
                if self._peek('","', '"NoSymbol"', '"SetMods"', '"LockGroup"', 'UNI', 'UNI3', 'HEX', 'SYM', context=_context) == '","':
                    self._scan('","', context=_context)
                sym = self.sym(_context)
                if sym: k.append(sym)
        self._scan('"\\\\]"', context=_context)

    def sym(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'sym', [])
        _token = self._peek('"NoSymbol"', '"SetMods"', '"LockGroup"', 'UNI', 'UNI3', 'HEX', 'SYM', context=_context)
        if _token == '"NoSymbol"':
            self._scan('"NoSymbol"', context=_context)
            return None
        elif _token == '"SetMods"':
            self._scan('"SetMods"', context=_context)
            self._scan('"\\\\("', context=_context)
            SYM = self._scan('SYM', context=_context)
            self._scan('"="', context=_context)
            SYM = self._scan('SYM', context=_context)
            self._scan('"\\\\)"', context=_context)
            return None
        elif _token == '"LockGroup"':
            self._scan('"LockGroup"', context=_context)
            self._scan('"\\\\("', context=_context)
            SYM = self._scan('SYM', context=_context)
            self._scan('"="', context=_context)
            NUM = self._scan('NUM', context=_context)
            self._scan('"\\\\)"', context=_context)
            return None
        elif _token == 'UNI':
            UNI = self._scan('UNI', context=_context)
            return eval('u"\\u%s"' % (UNI[1:]),{},{})
        elif _token == 'UNI3':
            UNI3 = self._scan('UNI3', context=_context)
            return eval('u"\\u0%s"' % (UNI3[1:]),{},{})
        elif _token == 'HEX':
            HEX = self._scan('HEX', context=_context)
            return eval('u"\\U%08x"' % (eval(HEX)-0x1000000,),{},{})
        else: # == 'SYM'
            SYM = self._scan('SYM', context=_context)
            return lookup_symbol(SYM)


def parse(rule, text):
    P = x11map(x11mapScanner(text))
    return runtime.wrap_error_reporter(P, rule)

# End -- grammar generated by Yapps


