"""
This code implements a parser for Linux keyboard maps.
"""

from keymapper.parse.linuxsyms import lookup_symbol,SymbolNotFound
from os.path import exists,join,curdir
from gzip import GzipFile
from keymapper.keymap import Keymap,Modifier
import os

dirs = [
]
def add_dir(d):
	"""Helper to add a directory to the search path"""
	dirs.append(d)

class NoFileFound(Exception):
	pass
class FileProblem(Exception):
	pass
class TooManySyms(Exception):
	pass

def get_file(name):
	"""Find a keymap file"""
	global dirs
	if not dirs:
		dirs=(curdir,)
	for end in ("",".kmap",".kmap.gz",".inc",".inc.gz", ".gz"):
		for dir in dirs:
			n = os.path.join(dir,name+end)
			if exists(n):
				if n.endswith(".gz"):
					f = GzipFile(n,"rb")
				else:
					f = open(n,"r")
				return f

	raise NoFileFound(name)

class ConsoleKeymap(Keymap):
	"""submap with interesting features for generating Linux console maps"""
	def __init__(self,*a,**k):
		Keymap.__init__(self,*a,**k)
		self.map_vector = []

	def add_map_vector(self,a,b=None):
		"""add to the list of keymaps"""
		if b is None: b=a
		self.map_vector.append((a,b))
	
	def get_map_vector(self):
		v = self.map_vector
		if not v: v = ((0,255),)

		for a,b in v:
			for i in range(a,b+1):
				yield Modifier(bits=i)

	def compose(self,c1,c2,result):
		pass
	
	def add_more(self, rvalue, code, keys):
		try:
			keys = keys.next()
		except StopIteration:
			raise TooManySyms(code,rvalue)
		else:
			# print rvalue, code, keys
			self.add(rvalue, code, keys)


def parse_file(name, with_alt=True):
	f = get_file(name)
	keymap = ConsoleKeymap(name, with_alt=with_alt)
	parser = linuxmap(linuxmapScanner("",file=f,filename=name))

	try:
		parser.map(keymap)
	except SymbolNotFound,e:
		print >>sys.stderr, "Keysym '%s' not found" % e.args
		print >>sys.stderr, parser._scanner
		raise FileProblem(name)
	except runtime.SyntaxError, e:
		runtime.print_error(e, parser._scanner, max_ctx=1)
		raise FileProblem(name)
	except runtime.NoMoreTokens:
		print >>sys.stderr, 'Could not complete parsing; stopped around here:'
		print >>sys.stderr, parser._scanner
		raise FileProblem(name)

	return keymap


# Begin -- grammar generated by Yapps
import sys, re
from yapps import runtime

class linuxmapScanner(runtime.Scanner):
    patterns = [
        ('"-"', re.compile('-')),
        ('"include"', re.compile('include')),
        ('[ \r\t]+', re.compile('[ \r\t]+')),
        ('\\\\\\n', re.compile('\\\\\\n')),
        ('(#|!).*', re.compile('(#|!).*')),
        ('EOF', re.compile('$')),
        ('EOL', re.compile('\n')),
        ('BLANK', re.compile('[ \t]')),
        ('INCLUDE', re.compile('include[ \t]*')),
        ('_DECIMAL', re.compile('[1-9][0-9]*')),
        ('_OCTAL', re.compile('0[0-7]*')),
        ('_HEX', re.compile('0[xX][0-9a-fA-F]+')),
        ('_UNICODE', re.compile('U\\+([0-9a-fA-F]){4}')),
        ('LITERAL', re.compile('[a-zA-Z][a-zA-Z_0-9]*')),
        ('OCTA', re.compile('([0-7]){1,3}')),
        ('CHARSET', re.compile('charset|Charset|CharSet|CHARSET')),
        ('KEYMAPS', re.compile('keymaps|Keymaps|KeyMaps|KEYMAPS')),
        ('KEYCODE', re.compile('keycode|Keycode|KeyCode|KEYCODE')),
        ('STRING', re.compile('string|String|STRING')),
        ('EQUALS', re.compile('=')),
        ('COMMA', re.compile(',')),
        ('PLAIN', re.compile('plain|Plain|PLAIN')),
        ('SHIFT', re.compile('shift|Shift|SHIFT')),
        ('CONTROL', re.compile('control|Control|CONTROL')),
        ('ALT', re.compile('alt|Alt|ALT')),
        ('ALTGR', re.compile('altgr|Altgr|AltGr|ALTGR')),
        ('SHIFTL', re.compile('shiftl|ShiftL|SHIFTL')),
        ('SHIFTR', re.compile('shiftr|ShiftR|SHIFTR')),
        ('CTRLL', re.compile('ctrll|CtrlL|CTRLL')),
        ('CTRLR', re.compile('ctrlr|CtrlR|CTRLR')),
        ('ALT_IS_META', re.compile('[aA][lL][tT][-_][iI][sS][-_][mM][eE][tT][aA]')),
        ('STRINGS', re.compile('strings|Strings|STRINGS')),
        ('COMPOSE', re.compile('compose|Compose|COMPOSE')),
        ('AS', re.compile('as|As|AS')),
        ('USUAL', re.compile('usual|Usual|USUAL')),
        ('FOR', re.compile('for|For|FOR')),
        ('ON', re.compile('on|On|ON')),
        ('TO', re.compile('to|To|TO')),
        ('_CHAR1', re.compile("'([0-7]){1,3}'")),
        ('_CHAR2', re.compile("'\\\\.'")),
        ('_CHAR3', re.compile("'.'")),
        ('PLUS', re.compile('\\+')),
        ('_STRLITERAL', re.compile('\'([^\\n\'\\\\]|\\\\.)*\'|"([^\\n"\\\\]|\\\\.)*"')),
    ]
    def __init__(self, str,*args,**kw):
        runtime.Scanner.__init__(self,None,{'(#|!).*':None,'[ \r\t]+':None,'\\\\\\n':None,},str,*args,**kw)

class linuxmap(runtime.Parser):
    Context = runtime.Context
    def DECIMAL(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'DECIMAL', [])
        _DECIMAL = self._scan('_DECIMAL', context=_context)
        return int(_DECIMAL)

    def OCTAL(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'OCTAL', [])
        _OCTAL = self._scan('_OCTAL', context=_context)
        return eval(_OCTAL,{},{})

    def HEX(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'HEX', [])
        _HEX = self._scan('_HEX', context=_context)
        return eval(_HEX,{},{})

    def UNICODE(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'UNICODE', [])
        _UNICODE = self._scan('_UNICODE', context=_context)
        return eval("u'\\u%s'" % _UNICODE[2:],{},{})

    def CCHAR(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'CCHAR', [])
        _token = self._peek('_CHAR1', '_CHAR2', '_CHAR3', context=_context)
        if _token == '_CHAR1':
            _CHAR1 = self._scan('_CHAR1', context=_context)
            return eval(_CHAR1,{},{})
        elif _token == '_CHAR2':
            _CHAR2 = self._scan('_CHAR2', context=_context)
            return _CHAR2[2]
        else: # == '_CHAR3'
            _CHAR3 = self._scan('_CHAR3', context=_context)
            return _CHAR3[1]

    def NUMBER(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'NUMBER', [])
        _token = self._peek('_DECIMAL', '_OCTAL', '_HEX', context=_context)
        if _token == '_DECIMAL':
            DECIMAL = self.DECIMAL(_context)
            return DECIMAL
        elif _token == '_OCTAL':
            OCTAL = self.OCTAL(_context)
            return OCTAL
        else: # == '_HEX'
            HEX = self.HEX(_context)
            return HEX

    def STRLITERAL(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'STRLITERAL', [])
        _STRLITERAL = self._scan('_STRLITERAL', context=_context)
        return eval(_STRLITERAL,{},{})

    def map(self, data, _parent=None):
        _context = self.Context(_parent, self._scanner, 'map', [data])
        self.data = data
        while self._peek('EOF', 'EOL', 'CHARSET', 'ALT_IS_META', 'STRINGS', 'KEYMAPS', 'KEYCODE', 'PLAIN', 'STRING', 'COMPOSE', '"include"', 'COMMA', 'SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', context=_context) not in ['EOF', 'COMMA']:
            line = self.line(_context)
        EOF = self._scan('EOF', context=_context)

    def line(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'line', [])
        _token = self._peek('EOL', 'CHARSET', 'ALT_IS_META', 'STRINGS', 'KEYMAPS', 'KEYCODE', 'PLAIN', 'STRING', 'COMPOSE', '"include"', 'SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', context=_context)
        if _token == 'EOL':
            EOL = self._scan('EOL', context=_context)
        elif _token == 'CHARSET':
            charsetline = self.charsetline(_context)
        elif _token == 'ALT_IS_META':
            altismetaline = self.altismetaline(_context)
        elif _token == 'STRINGS':
            usualstringsline = self.usualstringsline(_context)
        elif _token == 'KEYMAPS':
            keymapline = self.keymapline(_context)
        elif _token == 'KEYCODE':
            fullline = self.fullline(_context)
        elif _token not in ['STRING', 'COMPOSE', '"include"']:
            singleline = self.singleline(_context)
        elif _token == 'STRING':
            strline = self.strline(_context)
        elif _token == 'COMPOSE':
            compline = self.compline(_context)
        else: # == '"include"'
            include = self.include(_context)

    def include(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'include', [])
        self._scan('"include"', context=_context)
        STRLITERAL = self.STRLITERAL(_context)
        self._scanner.stack_input(file=get_file(STRLITERAL), filename=STRLITERAL)

    def charsetline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'charsetline', [])
        CHARSET = self._scan('CHARSET', context=_context)
        STRLITERAL = self.STRLITERAL(_context)
        EOL = self._scan('EOL', context=_context)

    def altismetaline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'altismetaline', [])
        ALT_IS_META = self._scan('ALT_IS_META', context=_context)
        EOL = self._scan('EOL', context=_context)

    def usualstringsline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'usualstringsline', [])
        STRINGS = self._scan('STRINGS', context=_context)
        AS = self._scan('AS', context=_context)
        USUAL = self._scan('USUAL', context=_context)
        EOL = self._scan('EOL', context=_context)

    def compline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'compline', [])
        COMPOSE = self._scan('COMPOSE', context=_context)
        _token = self._peek('AS', '_CHAR1', '_CHAR2', '_CHAR3', context=_context)
        if _token == 'AS':
            usualcomposeline = self.usualcomposeline(_context)
        else: # in ['_CHAR1', '_CHAR2', '_CHAR3']
            composeline = self.composeline(_context)

    def usualcomposeline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'usualcomposeline', [])
        AS = self._scan('AS', context=_context)
        USUAL = self._scan('USUAL', context=_context)
        if self._peek('FOR', 'EOL', 'EOF', 'COMMA', 'CHARSET', 'ALT_IS_META', 'STRINGS', 'KEYMAPS', 'KEYCODE', 'PLAIN', 'STRING', 'COMPOSE', '"include"', 'SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', context=_context) == 'FOR':
            FOR = self._scan('FOR', context=_context)
            STRLITERAL = self.STRLITERAL(_context)
        EOL = self._scan('EOL', context=_context)

    def keymapline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'keymapline', [])
        KEYMAPS = self._scan('KEYMAPS', context=_context)
        range = self.range(_context)
        while self._peek('COMMA', 'EOF', 'EOL', 'CHARSET', 'ALT_IS_META', 'STRINGS', 'KEYMAPS', 'KEYCODE', 'PLAIN', 'STRING', 'COMPOSE', '"include"', 'SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', context=_context) == 'COMMA':
            COMMA = self._scan('COMMA', context=_context)
            range = self.range(_context)

    def range(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'range', [])
        NUMBER = self.NUMBER(_context)
        n1 = NUMBER
        _token = self._peek('"-"', 'COMMA', 'EOF', 'EOL', 'CHARSET', 'ALT_IS_META', 'STRINGS', 'KEYMAPS', 'KEYCODE', 'PLAIN', 'STRING', 'COMPOSE', '"include"', 'SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', context=_context)
        if _token == '"-"':
            self._scan('"-"', context=_context)
            NUMBER = self.NUMBER(_context)
            self.data.add_map_vector(n1,NUMBER)
        else:
            self.data.add_map_vector(n1)

    def strline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'strline', [])
        STRING = self._scan('STRING', context=_context)
        LITERAL = self._scan('LITERAL', context=_context)
        EQUALS = self._scan('EQUALS', context=_context)
        STRLITERAL = self.STRLITERAL(_context)
        EOL = self._scan('EOL', context=_context)

    def composeline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'composeline', [])
        CCHAR = self.CCHAR(_context)
        c1 = CCHAR
        CCHAR = self.CCHAR(_context)
        c2 = CCHAR
        TO = self._scan('TO', context=_context)
        _token = self._peek('_CHAR1', '_CHAR2', '_CHAR3', 'PLUS', '_DECIMAL', '_OCTAL', '_HEX', '_UNICODE', 'LITERAL', context=_context)
        if _token in ['_CHAR1', '_CHAR2', '_CHAR3']:
            CCHAR = self.CCHAR(_context)
            self.data.compose(c1, c2, CCHAR);
        else:
            rvalue = self.rvalue(_context)
            self.data.compose(c1, c2, rvalue);
        EOL = self._scan('EOL', context=_context)

    def singleline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'singleline', [])
        _token = self._peek('PLAIN', 'SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', context=_context)
        if _token != 'PLAIN':
            mods = 0;
            modifiers = self.modifiers(_context)
            KEYCODE = self._scan('KEYCODE', context=_context)
            NUMBER = self.NUMBER(_context)
            EQUALS = self._scan('EQUALS', context=_context)
            rvalue = self.rvalue(_context)
            EOL = self._scan('EOL', context=_context)
            self.data.add(rvalue, NUMBER, Modifier(modifiers));
        else: # == 'PLAIN'
            PLAIN = self._scan('PLAIN', context=_context)
            KEYCODE = self._scan('KEYCODE', context=_context)
            NUMBER = self.NUMBER(_context)
            EQUALS = self._scan('EQUALS', context=_context)
            rvalue = self.rvalue(_context)
            EOL = self._scan('EOL', context=_context)
            self.data.add(rvalue, NUMBER, Modifier(()));

    def modifiers(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'modifiers', [])
        vals = []
        while 1:
            modifier = self.modifier(vals, _context)
            if self._peek('SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', 'KEYCODE', context=_context) not in ['SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR']: break
        return vals

    def modifier(self, vals, _parent=None):
        _context = self.Context(_parent, self._scanner, 'modifier', [vals])
        mod = self.mod(_context)
        vals.append(mod)

    def mod(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'mod', [])
        _token = self._peek('SHIFT', 'CONTROL', 'ALT', 'ALTGR', 'SHIFTL', 'SHIFTR', 'CTRLL', 'CTRLR', context=_context)
        if _token == 'SHIFT':
            SHIFT = self._scan('SHIFT', context=_context)
            return "shift"
        elif _token == 'CONTROL':
            CONTROL = self._scan('CONTROL', context=_context)
            return "control"
        elif _token == 'ALT':
            ALT = self._scan('ALT', context=_context)
            return "alt"
        elif _token == 'ALTGR':
            ALTGR = self._scan('ALTGR', context=_context)
            return "altgr"
        elif _token == 'SHIFTL':
            SHIFTL = self._scan('SHIFTL', context=_context)
            return "shiftL"
        elif _token == 'SHIFTR':
            SHIFTR = self._scan('SHIFTR', context=_context)
            return "shiftR"
        elif _token == 'CTRLL':
            CTRLL = self._scan('CTRLL', context=_context)
            return "controlL"
        else: # == 'CTRLR'
            CTRLR = self._scan('CTRLR', context=_context)
            return "controlR"

    def fullline(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'fullline', [])
        KEYCODE = self._scan('KEYCODE', context=_context)
        NUMBER = self.NUMBER(_context)
        EQUALS = self._scan('EQUALS', context=_context)
        rvalue0 = self.rvalue0(NUMBER, _context)
        EOL = self._scan('EOL', context=_context)

    def rvalue0(self, code, _parent=None):
        _context = self.Context(_parent, self._scanner, 'rvalue0', [code])
        keys = self.data.get_map_vector()
        while self._peek('EOL', 'PLUS', '_DECIMAL', '_OCTAL', '_HEX', '_UNICODE', 'LITERAL', context=_context) != 'EOL':
            rvalue = self.rvalue(_context)
            self.data.add_more(rvalue, code, keys)

    def rvalue(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'rvalue', [])
        _token = self._peek('PLUS', '_DECIMAL', '_OCTAL', '_HEX', '_UNICODE', 'LITERAL', context=_context)
        if _token != 'PLUS':
            rvalue2 = self.rvalue2(_context)
            return rvalue2
        else: # == 'PLUS'
            PLUS = self._scan('PLUS', context=_context)
            rvalue2 = self.rvalue2(_context)
            return rvalue2

    def rvalue2(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'rvalue2', [])
        _token = self._peek('_DECIMAL', '_OCTAL', '_HEX', '_UNICODE', 'LITERAL', context=_context)
        if _token not in ['_UNICODE', 'LITERAL']:
            NUMBER = self.NUMBER(_context)
            return eval("u'\\u%04x'" % NUMBER, {},{})
        elif _token == '_UNICODE':
            UNICODE = self.UNICODE(_context)
            return UNICODE
        else: # == 'LITERAL'
            LITERAL = self._scan('LITERAL', context=_context)
            return lookup_symbol(LITERAL)


def parse(rule, text):
    P = linuxmap(linuxmapScanner(text))
    return runtime.wrap_error_reporter(P, rule)

# End -- grammar generated by Yapps


